import pandas as pd
import requests
import json
import geopandas as gpd

class MarcoGeoestadistico:
    def __init__(self):
        self.__liga_base = 'https://gaia.inegi.org.mx/wscatgeo/'
        
    # uiliza la liga/url para obtener el geojson y pasarlo a un geodataframe
    def __obtener_df(self, liga, as_gdf):
        req = requests.get(liga)
        assert req.status_code == 200, 'No se encontró información con las parámetros especificados. \n Liga: {}'.format(liga)
        data =  json.loads(req.text)
        if as_gdf: 
            try: 
                gdf = gpd.GeoDataFrame.from_features(data["features"])
                gdf.crs = 'EPSG:4326'
                return gdf
            except: raise Exception('No se encontró información con las parámetros especificados. \n Liga: {}'.format(liga))
        else: 
            try: return pd.DataFrame.from_dict(data['datos'])
            except: raise Exception('No se encontró información con las parámetros especificados. \n Liga: {}'.format(liga))
    
    # construye la liga y la lista de entidades
    # se tiene que ver la mejor fórma de usar esta función para entidades más desagregadas:
    # a cada nivel de agregación se añade una variable extra y complejiza más la consulta
    def __liga_y_areas(self, nombres, claves, tipo_area, as_gdf):
        if as_gdf: liga = '{}geo/'.format(self.__liga_base)
        else: liga =  self.__liga_base
        
        if nombres: 
            liga = '{}{}/buscar/'.format(liga, tipo_area)
            areas_geoestadisticas = nombres
        else: 
            liga = '{}{}/'.format(liga, tipo_area)
            areas_geoestadisticas = claves
        
        if isinstance(areas_geoestadisticas, str): areas_geoestadisticas = [areas_geoestadisticas]
        
        return liga, areas_geoestadisticas
    
    def __desconcatenador_de_claves(self, claves_concatenadas):
        entidades = list()
        municipios = list()
        localidades = list()
        if isinstance(claves_concatenadas, str): claves_concatenadas = [claves_concatenadas]
        for clave in claves_concatenadas:
            if len(clave[:2]) > 0: entidades.append(clave[:2])
            else: entidades.append(None)# aquí sería agregar un raise exception

            if len(clave[2:5]) > 0: municipios.append(clave[2:5])
            else: municipios.append(None)

            if len(clave[5:9]) > 0: localidades.append(clave[5:9])
            else: localidades.append(None)
            
        return entidades, municipios, localidades
    
    # para el caso de localidades lo mejor es concatenar en una sola clave en vez de separar con diagonales,
    # para eso se agrego la variable separadas.
    def __obtener_claves(self, entidades, municipios, localidades, agebs, claves_concatenadas, separadas = True):
        if claves_concatenadas: entidades, municipios, localidades = self.__desconcatenador_de_claves(claves_concatenadas)
        niveles = [entidades, municipios, localidades, agebs]
        if separadas: s = '{}/'
        else: s = '{}'
        clave = ''
        for nivel in niveles:
            if nivel is None: return clave
            elif isinstance(nivel, str): clave = clave + s.format(nivel)
            elif isinstance(nivel, list) and isinstance(clave, str): clave = [clave + s.format(n) for n in nivel] # más bien sería claves pero tiene iun punto seguir nombrándola clave
            elif isinstance(nivel, list) and isinstance(clave, list): clave = [clave[i]  + s.format(nivel[i]) if nivel[i] else clave[i] for i in range(len(nivel))]
            
        return clave
                
    def __obtener_consulta(self, liga, areas_geoestadisticas, as_gdf, ambito = None):
        if ambito not in ['rural','urbano', 'r','u',None]: raise Exception('El ámbito proporcionado no es correcto.') 
        dfs = list()
        if areas_geoestadisticas: 
            for area in areas_geoestadisticas: 
                liga_final = liga + area
                if ambito: liga_final = liga_final + ambito.capitalize()[0]
                df = self.__obtener_df(liga_final, as_gdf)
                dfs.append(df)
            c_df = pd.concat(dfs, axis = 0)
            return c_df.reset_index(drop = True)
        else:
            return self.__obtener_df(liga, as_gdf)
    
    def Entidades(self, 
                  entidades: 'str|list' = None, 
                  nombres: 'str|list' = None, 
                  as_geodf: bool = True):
        '''
        Obtiene el DataFrame con las áreas geoestadísticas estatales.
        
        Si no se especifíca un estado o lista de estados la función regresa un GeoDataFrame con todas las Entidades Federativas.
        
        Parámetros:
        -----------
        entidades: str/list. Clave(s) AGEE de dos dígitos (01 a 32) de las entidades federativas a buscar.
        nombres: str/list. Nombre(s) de los estados a buscar.Si se define este parámetro obtiene las áreas geoestadísticas estatales cuyo nombre contenga el texto proporcionado sin necesidad de proporcionar las claves de las entidades. No distingue entre mayúsculas y minúsculas pero sí considera acentos.
        as_geodf: bool. Si el valor es verdadero regresa un GeoDataFrame para facilitar el análisis espacial. 
        -----------
        
        Regresa un Dataframe o GeoDataFrame con las áreas seleccionadas.
        
        Para más información consultar: https://www.inegi.org.mx/servicios/catalogoUnico.html
        
        '''
        liga, areas_geoestadisticas = self.__liga_y_areas(nombres, entidades, 'mgee', as_geodf)
        df = self.__obtener_consulta(liga, areas_geoestadisticas, as_geodf)
        df.iloc[:, -4:] = df.iloc[:, -4:].astype(int)
        return df
    
    # esta requiere llamar la función concatenadora
    def Municipios(self, 
                   entidades: 'str|list' = None, 
                   municipios: 'str|list' = None, 
                   nombres: 'str|list' = None, 
                   claves_concatenadas: 'str|list' = None, 
                   as_geodf: bool = True):
        '''
        Obtiene el DataFrame con las áreas geoestadísticas municipales.
        
        Si no se especifícan los parámetros regresa un DataFrame con todos los municipios de México.
        
        Nota: si se pasa una lista de entidades no se puede definir un municipio en específico a buscar. En este caso se regresan todos los municipios de la lista de entidades.
        
        Parámetros:
        -----------
        entidades: str/list. Clave(s) AGEE de dos dígitos (01 a 32) de las entidades federativas a buscar.
        municipios: str/list. Clave(s) AGEM de tres dígitos de los municipios a buscar (ej. 001).
        nombres: str/list. Nombre(s) de los municipios a buscar. Si se define este parámetro obtiene las áreas geoestadísticas municipales cuyo nombre contenga el texto proporcionado sin necesidad de proporcionar las claves de los municipios. No distingue entre mayúsculas y minúsculas pero sí considera acentos.
        claves_concatendas: str/list. Clave(s) concatenada con los niveles de agregación espacial.
        as_geodf: bool. Si el valor es verdadero regresa un GeoDataFrame para facilitar el análisis espacial. 
        -----------
        
        Regresa un DataFrame o GeoDataFrame con las áreas seleccionadas.
        
        Para más información consultar: https://www.inegi.org.mx/servicios/catalogoUnico.html
        
        '''
        claves = self.__obtener_claves(entidades, municipios, None, None, claves_concatenadas)
        liga, areas_geoestadisticas = self.__liga_y_areas(nombres, claves, 'mgem', as_geodf)
        df = self.__obtener_consulta(liga, areas_geoestadisticas, as_geodf)
        df.iloc[:, -4:] = df.iloc[:, -4:].astype(int)
        return df
    
    def LocalidadesAmanzanadas(self, 
                               entidades: 'str|list' = None, 
                               municipios: 'str|list' = None, 
                               localidades: 'str|list' = None, 
                               nombres: 'str|list' = None, 
                               claves_concatenadas: 'str|list' = None, 
                               ambito: str = None, 
                               as_geodf: bool = True):
        '''
        Obtiene el DataFrame con las áreas geoestadísticas a nivel localidad amanzanada.
        
        Si no se especifícan los parámetros regresa un DataFrame con todas las localidades de México. 
        
        Nota: si se pasa una lista de áreas con nivel de agregación menor a la localidad no se puede definir una localidad en específico a buscar. En este caso se regresan todos las localidades de los niveles previos. 
        
        Parámetros:
        -----------
        entidades: str/list. Clave(s) AGEE de dos dígitos (01 a 32) de las entidades federativas a buscar.
        municipios: str/list. Clave(s) AGEM de tres dígitos de los municipios a buscar (ej. 001)
        localidades: str/list. Clave(s) de cuatro dígitos de las localidades a buscar (ej. 0001 )
        nombres: str/list. Nombre(s) de las localidades a buscar. Si se define este parámetro obtiene las localidades amanzanadas cuyo nombre contenga el texto proporcionado sin necesidad de proporcionar las claves de las localidades. No distingue entre mayúsculas y minúsculas pero sí considera acentos.
        claves_concatendas: str/list. Clave(s) concatenada con los niveles de agregación espacial.
        ambito: str. ['urbano'|'rural'] Define el ambito de las localidades. Si se define un ámbito no se puede definir una localidad en específico y se debe definir tanto entidad como municipio.
        as_geodf: bool. Si el valor es verdadero regresa un GeoDataFrame para facilitar el análisis espacial. 
        -----------
        
        Regresa un DataFrame o GeoDataFrame con las áreas seleccionadas.
        
        Para más información consultar: https://www.inegi.org.mx/servicios/catalogoUnico.html
        
        '''
        if ambito: claves = self.__obtener_claves(entidades, municipios, localidades, None, claves_concatenadas)
        else: claves = self.__obtener_claves(entidades, municipios, localidades, None, claves_concatenadas, False)
        liga, areas_geoestadisticas = self.__liga_y_areas(nombres, claves, 'localidades/pol', True)
        df = self.__obtener_consulta(liga, areas_geoestadisticas, True, ambito)
        if as_geodf is False: df = pd.DataFrame(df.drop('geometry',axis=1))
        try: df.iloc[:, -5:-3] = df.iloc[:, -5:-3].astype(int)
        except: pass
        return df
    
    def LocalidadesRuralesPuntuales(self, 
                                    entidades: 'str|list' = None, 
                                    municipios: 'str|list' = None, 
                                    localidades: 'str|list' = None, 
                                    nombres: 'str|list' = None, 
                                    claves_concatenadas: 'str|list' = None, 
                                    as_geodf: bool = True):
        '''
        Obtiene el DataFrame con las áreas geoestadísticas a nivel localidad rural puntual.
        
        Si no se especifícan los parámetros regresa un DataFrame con todas las localidades rurales puntuales de México. 
        
        Nota: si se pasa una lista de áreas con nivel de agregación menor a la localidad no se puede definir una localidad en específico a buscar. En este caso se regresan todos las localidades de los niveles previos. 
        
        Parámetros:
        -----------
        entidades: str/list. Clave(s) AGEE de dos dígitos (01 a 32) de las entidades federativas a buscar.
        municipios: str/list. Clave(s) AGEM de tres dígitos de los municipios a buscar (ej. 001)
        loacalidades: str/list. Clave(s) de cuatro dígitos de las localidades a buscar (ej. 0001 )
        nombres: str/list. Nombre(s) de las localidades a buscar. Si se define este parámetro obtiene las localidades amanzanadas cuyo nombre contenga el texto proporcionado sin necesidad de proporcionar las claves de las localidades. No distingue entre mayúsculas y minúsculas pero sí considera acentos.
        claves_concatendas: str/list. Clave(s) concatenada con los niveles de agregación espacial.
        as_geodf: bool. Si el valor es verdadero regresa un GeoDataFrame para facilitar el análisis espacial. 
        -----------
        
        Regresa un DataFrame o GeoDataFrame con las áreas seleccionadas.
        
        Para más información consultar: https://www.inegi.org.mx/servicios/catalogoUnico.html
        
        '''
        claves = self.__obtener_claves(entidades, municipios, localidades, None, claves_concatenadas, False)
        liga, areas_geoestadisticas = self.__liga_y_areas(nombres, claves, 'localidades/ruralespto', True)
        df = self.__obtener_consulta(liga, areas_geoestadisticas, True)
        if as_geodf is False: df = pd.DataFrame(df.drop('geometry',axis=1))
        try: df.iloc[:, -5:-3] = df.iloc[:, -5:-3].astype(int)
        except: pass
        return df
    
    def AGEBs(self, 
              entidades: 'str|list' = None, 
              municipios: 'str|list' = None, 
              localidades: 'str|list' = None, 
              agebs: 'str|list' = None, 
              claves_concatenadas: 'str|list' = None, 
              ambito: str = 'urbano', 
              as_geodf: bool = True):
        '''
        Obtiene el DataFrame con las áreas geoestadísticas a nivel Área Geoestadística Básica. 
        
        Nota: si se pasa una lista de áreas con nivel de agregación menor al AGEB no se puede definir una localidad en específico a buscar. En este caso se regresan todos los AGEBs de los niveles previos. 
        
        Parámetros:
        -----------
        entidades: str/list. Clave(s) AGEE de dos dígitos (01 a 32) de las entidades federativas a buscar.
        municipios: str/list. Clave(s) AGEM de tres dígitos de los municipios a buscar (ej. 001).
        loacalidades: str/list. Clave(s) de cuatro dígitos de las localidades a buscar (ej. 0001).
        agebs: str/list. Clave(s) de cuatro dígitos con las AGEBs a buscar (ej. 2000).
        claves_concatendas: str/list. Clave(s) concatenada con los niveles de agregación espacial.
        ambito: str. ['urbano'|'rural'] Define el ambito de las AGEBs. A diferencia de las localidades amanzanadas siempre se debe especificar el ambitos de las AGEBsa buscar. Por default son urbanas.
        as_geodf: bool. Si el valor es verdadero regresa un GeoDataFrame para facilitar el análisis espacial. 
        -----------
        
        Regresa un DataFrame o GeoDataFrame con las áreas seleccionadas.
        
        Para más información consultar: https://www.inegi.org.mx/servicios/catalogoUnico.html
        
        '''
        tipo_area = 'ageb{}'.format(ambito[0].lower())
        if agebs: claves = self.__obtener_claves(entidades, municipios, localidades, agebs, claves_concatenadas, False)
        else: claves = self.__obtener_claves(entidades, municipios, localidades, agebs, claves_concatenadas)
        if claves_concatenadas: claves = [self.__obtener_claves(claves_concatenadas=clave)[0] if len(clave) <= 5 else clave for clave in claves]
        liga, areas_geoestadisticas = self.__liga_y_areas(None, claves, tipo_area, as_geodf)
        df = self.__obtener_consulta(liga, areas_geoestadisticas, as_geodf)
        try: df.iloc[:, -5:-1] = df.iloc[:, -5:-1].astype(int)
        except: pass
        return df
        
    
    def Manzanas(self, 
                 entidades: 'str|list' = None,
                 municipios: 'str|list' = None, 
                 localidades: 'str|list' = None, 
                 claves_concatenadas: 'str|list' = None, 
                 ambito: str = None, 
                 as_geodf: bool = True):
        '''
        Obtiene el DataFrame con las áreas geoestadísticas a nivel manzana. 
        
        Nota: si se pasa una lista de áreas con nivel de agregación menor a la localidad no se puede definir una localidad en específico a buscar. En este caso se regresan todos las localidades de los niveles previos. 
        
        Tampoco se puede difinir un nivel de agregación sin definir el nivel previo. No se puede definir un municipio sin definir una entidad, etc.
        
        Parámetros:
        -----------
        entidades: str/list. Clave(s) AGEE de dos dígitos (01 a 32) de las entidades federativas a buscar.
        municipios: str/list. Clave(s) AGEM de tres dígitos de los municipios a buscar (ej. 001)
        localidades: str/list. Clave(s) de cuatro dígitos de las localidades a buscar (ej. 0001)
        claves_concatendas: str/list. Clave(s) concatenada con los niveles de agregación espacial.
        ambito: str. ['urbano'|'rural'] Define el ambito de las manzanas. Si se define un ámbito no se puede definir una localidad en específico y se debe definir tanto entidad como municipio.
        as_geodf: bool. Si el valor es verdadero regresa un GeoDataFrame para facilitar el análisis espacial. 
        -----------
        
        Regresa un DataFrame o GeoDataFrame con las áreas seleccionadas.
        
        Para más información consultar: https://www.inegi.org.mx/servicios/catalogoUnico.html
        
        '''
        claves = self.__obtener_claves(entidades, municipios, localidades, None, claves_concatenadas)
        liga, areas_geoestadisticas = self.__liga_y_areas(None, claves, 'mza', as_geodf)
        df = self.__obtener_consulta(liga, areas_geoestadisticas, as_geodf, ambito)
        try: df.loc[:, ["pobtot", "pobmas", "pobfem", "tvivhab"]] = df.loc[:, ["pobtot", "pobmas", "pobfem", "tvivhab"]].astype(float)
        except: pass
        return df
    
    def Vialidades(self, 
                   entidades: 'str|list' = None, 
                   municipios: 'str|list' = None, 
                   localidades: 'str|list' = None, 
                   claves_concatenadas: 'str|list' = None, 
                   as_geodf: bool = True):
        '''
        Obtiene el DataFrame con las áreas geoestadísticas a nivel vialidad.
        
        Nota: si se pasa una lista de áreas con nivel de agregación menor a la localidad no se puede definir una localidad en específico a buscar. En este caso se regresan todos las localidades de los niveles previos. 
        
        Tampoco se puede difinir un nivel de agregación sin definir el nivel previo. No se puede definir un municipio sin definir una entidad, etc.
        
        Parámetros:
        -----------
        entidades: str/list. Clave(s) AGEE de dos dígitos (01 a 32) de las entidades federativas a buscar.
        municipios: str/list. Clave(s) AGEM de tres dígitos de los municipios a buscar (ej. 001)
        localidades: str/list. Clave(s) de cuatro dígitos de las localidades a buscar (ej. 0001)
        claves_concatendas: str/list. Clave(s) concatenada con los niveles de agregación espacial.
        as_geodf: bool. Si el valor es verdadero regresa un GeoDataFrame para facilitar el análisis espacial. 
        -----------
        
        Regresa un DataFrame o GeoDataFrame con las áreas seleccionadas.
        
        Para más información consultar: https://www.inegi.org.mx/servicios/catalogoUnico.html
        
        '''
        if localidades or claves_concatenadas: 
            claves = self.__obtener_claves(entidades, municipios, localidades, None, claves_concatenadas, False)
            claves = [self.__obtener_claves(None, None, None, None, clave)[0] if len(clave) <= 5 else clave for clave in claves]
        else: claves = self.__obtener_claves(entidades, municipios, localidades, None, claves_concatenadas)
        liga, areas_geoestadisticas = self.__liga_y_areas(None, claves, 'vialidades', as_geodf)
        return self.__obtener_consulta(liga, areas_geoestadisticas, as_geodf)



import requests
import json
import numpy as np
import pandas as pd
from pandas import DataFrame
import geopandas as gpd
from shapely.geometry import shape

# Módulo dedicado al api de Ruteo del INEGI: https://www.inegi.org.mx/servicios/Ruteo/Default.html#token

# falta usar typing para definir los tipos que reciben los parámteros en las funciones de ruteo. 

class Ruteo:
    def __init__(self, token):
        self.__token = token
        self.__liga_base = 'https://gaia.inegi.org.mx/sakbe_v3.1/'
        
    def __obtener_consulta(self, funcion, params, as_geodf = True): # función base para realizar consulta y generar el geodataframe
        req = requests.post('https://gaia.inegi.org.mx/sakbe_v3.1/{}'.format(funcion), params=params)
        try: data = json.loads(req.text)['data']
        except: raise Exception("No se encontraro resultados para la búsqueda")
        if isinstance(data, dict): data = [data]
        df = pd.DataFrame.from_dict(data)
        if as_geodf:
            df['geometry'] = df['geojson'].apply(lambda x: shape(json.loads(x)))
            df = df.drop('geojson', axis=1)
            df = gpd.GeoDataFrame(df)
            df.crs = 'EPSG:4326'
        if 'punto_caseta' in df.columns:
            df.punto_caseta = df.punto_caseta.apply(lambda x: shape(json.loads(x)) if x else x)
        return df
        
    def BuscarDestino(self, 
                      busqueda: str, 
                      cantidad: int, 
                      proyeccion: str = 'GRS80'): 
        '''
        Permite buscar destinos como localidades urbanas y rurales, así como los sitios de interés que pueden ser instalaciones de servicios como aeropuertos, puertos, servicios médicos, centros educativos de nivel superior, así como sitios atractivos para el turismo como playas, cascadas, zonas arqueológicas, museos, pueblos mágicos, y más.
        
        Parámetros:
        -----------
        busqueda: str. Define el nombre o parte del destino que se desea encontrar. Se puede utilizar una coma para especificar la entidad federativa, p. e. “San Juan, Jalisco”.
        cantidad: int. Número de destinos que se desea obtener. 
        proyeccion: str. Define la proyección de los puntos resultantes. GRS80 para coordenadas geográficas y MERC para coordenadas Spherical Mercator. Por default será GRS80. 
        -----------
        
        Obtiene un GeoDataFrame con la información de los destinos registrados en la Red Nacional de Caminos. 

        Para más información consultar: https://www.inegi.org.mx/servicios/Ruteo/Default.html#token

        '''
        params = {'buscar':busqueda, 
                  'type':'json', 
                  'num':cantidad, 
                  'key':self.__token, 
                  'proj':proyeccion}
        return self.__obtener_consulta('buscadestino', params)
    
    def BuscarLinea(self, 
                    lat: float, 
                    lng: float, 
                    escala: int = 1_000_000, 
                    proyeccion: str = 'GRS80'):
        '''
        Obtiene un GeoDataFrame con la información de la línea registrada en la Red Nacional de Caminos más cercana a una coordenada. 
        
        Parámetros:
        -----------
        lat: float. Latitud de la coordenada.
        lng: float. Longitud de la coordenada.
        escala: int. Valor de la escala de visualización. Por default es 1,000,000.
        proyeccion: str. Define la proyección de los puntos resultantes. GRS80 para coordenadas geográficas y MERC para coordenadas Spherical Mercator. Por default será GRS80. 
        -----------

        Para más información consultar: https://www.inegi.org.mx/servicios/Ruteo/Default.html#token

        '''
        params = {'escala': escala, 
                  'type': 'json', 
                  'x': lng, 
                  'y': lat,
                  'key': self.__token, 
                  'proj': proyeccion}
        return self.__obtener_consulta('buscalinea', params)
    
    def __obtener_ruta(self, linea_inicial, linea_final, destino_inicial, destino_final, tipo_vehiculo, ruta, ejes_excedentes, saltar_lineas, proyeccion, funcion):
        rutas_validas = ['optima', 'libre', 'cuota']
        if ruta not in rutas_validas: raise ValueError("El tipo de ruta deber ser alguno de los siguientes {}".format(rutas_validas))
        if funcion == 'detalle_': funcion = '{}{}'.format(funcion, ruta[0])
        else: funcion = ruta
        params = {'type': 'json',
                  'key': self.__token, 
                  'proj': proyeccion,
                  'v': tipo_vehiculo,
                  'e': ejes_excedentes}
        
        if linea_inicial is not None:
            try: 
                n_params  = {'id_i': linea_inicial['id_routing_net'], 'source_i': linea_inicial['source'], 'target_i': linea_inicial['target']}
                params.update(n_params)
            except: raise Exception("Se deben proporcionar los parámterios id_routing_net, source y target de la línea inicial")
                                  
        if linea_final is not None:
            try: 
                n_params  = {'id_f': linea_final['id_routing_net'], 'source_f': linea_final['source'], 'target_f': linea_final['target']}
                params.update(n_params)
            except: raise Exception("Se deben proporcionar los parámterios id_routing_net, source y target de la línea final")
            
        if destino_inicial is not None: 
            if isinstance(destino_inicial, DataFrame): params['dest_i'] = destino_inicial['id_dest'].iloc[0]
            else: params['dest_i'] = destino_inicial
        if destino_final is not None:
            if isinstance(destino_final, DataFrame): params['dest_f'] = destino_final['id_dest'].iloc[0]
            else: params['dest_f'] = destino_final
        if saltar_lineas is not None: params['b'] = ','.join(saltar_lineas)
        
        return self.__obtener_consulta(funcion, params)
    
    def CalcularRuta(self, 
                     linea_inicial: 'DataFrame|dict' = None, 
                     linea_final: 'DataFrame|dict' = None, 
                     destino_inicial: 'DataFrame|dict|str|int' = None, 
                     destino_final: 'DataFrame|dict|str|int' = None, 
                     tipo_vehiculo: int = 0, 
                     ruta: str = 'optima', 
                     ejes_excedentes: int = 0, 
                     saltar_lineas: list = None, 
                     proyeccion: str = 'GRS80'):
        '''
        Obtiene un GeoDataFrame con la ruta calculada por Sistema de Ruteo de México y la Red Nacional de Caminos. Se puede obtener rutas de línea-línea, destino-destino, línea-destino y destino-linea. 
        
        Parámetros:
        -----------
        linea_inicial: DataFrame con las columnas id_routing_net, source y target de la línea inicial obtenido por la función BuscarLinea. También acepta un diccionario con la misma información.
        linea_final: DataFrame con las columnas id_routing_net, source y target de la línea final obtenido por la función BuscarLinea. También acepta un diccionario con la misma información.
        destino_inicial: DataFrame con la columna id_dest del destino inicial obtenido por la función BuscarDestino. También acepta el valor de id_dest en string o integer. 
        destino_final: DataFrame con la columna id_dest del destino final obtenido por la función BuscarDestino. También acepta el valor de id_dest en string o integer. 
        tipo_vehiculo: int. Clave con el tipo de vehículo. Por default es 0 que equivale a motocicleta mientras que 1 equivale a automóvil. Para consultar los demás valores se debe revisar la guía de desarrolladores en: https://www.inegi.org.mx/servicios/Ruteo/Default.html#token
        ruta: str. ['optima' | 'libre' | 'cuota']. Tipo de ruta que se desea obtener. 
        ejes_excedentes: int. Número de ejes excedentes del vehículo. Por default es 0 que equivale a ningíun eje excedente. Para consultar los demás valores se debe revisar la guía de desarrolladores en: https://www.inegi.org.mx/servicios/Ruteo/Default.html#token
        saltar_lineas: list. Lista con los id_routing_net de las líneas por las cuales la ruta no pasará por algún motivo. Por default es None.
        proyeccion: str. Define la proyección de los puntos resultantes. GRS80 para coordenadas geográficas y MERC para coordenadas Spherical Mercator. Por default será GRS80. 
        -----------

        Para más información consultar: https://www.inegi.org.mx/servicios/Ruteo/Default.html#token

        '''

        ruta = self.__obtener_ruta(linea_inicial, linea_final, destino_inicial, destino_final, tipo_vehiculo, ruta, ejes_excedentes, saltar_lineas, proyeccion, '')
        ruta['peaje'] = np.select([ruta.peaje == 'f'],[False], True)
        
        return ruta
    
    def DetalleRuta(self,
                    linea_inicial: 'DataFrame|dict' = None, 
                    linea_final: 'DataFrame|dict' = None, 
                    destino_inicial: 'DataFrame|dict|str|int' = None, 
                    destino_final: 'DataFrame|dict|str|int' = None, 
                    tipo_vehiculo: int = 0, 
                    ruta: str = 'optima', 
                    ejes_excedentes: int = 0, 
                    saltar_lineas: list = None, 
                    proyeccion: str = 'GRS80'):
        '''
        Obtiene un GeoDataFrame con los detalles de la ruta calculada por Sistema de Ruteo de México y la Red Nacional de Caminos. Se puede obtener rutas de línea-línea, destino-destino, línea-destino y destino-linea. 
        
        Parámetros:
        -----------
        linea_inicial: DataFrame con las columnas id_routing_net, source y target de la línea inicial obtenido por la función BuscarLinea. También acepta un diccionario con la misma información.
        linea_final: DataFrame con las columnas id_routing_net, source y target de la línea final obtenido por la función BuscarLinea. También acepta un diccionario con la misma información.
        destino_inicial: DataFrame con la columna id_dest del destino inicial obtenido por la función BuscarDestino. También acepta el valor de id_dest en string o integer. 
        destino_final: DataFrame con la columna id_dest del destino final obtenido por la función BuscarDestino. También acepta el valor de id_dest en string o integer. 
        tipo_vehiculo: int. Clave con el tipo de vehículo. Por default es 0 que equivale a motocicleta mientras que 1 equivale a automóvil. Para consultar los demás valores se debe revisar la guía de desarrolladores en: https://www.inegi.org.mx/servicios/Ruteo/Default.html#token
        ruta: str. ['optima' | 'libre' | 'cuota']. Tipo de ruta que se desea obtener. 
        ejes_excedentes: int. Número de ejes excedentes del vehículo. Por default es 0 que equivale a ningíun eje excedente. Para consultar los demás valores se debe revisar la guía de desarrolladores en: https://www.inegi.org.mx/servicios/Ruteo/Default.html#token
        saltar_lineas: list. Lista con los id_routing_net de las líneas por las cuales la ruta no pasará por algún motivo. Por default es None.
        proyeccion: str. Define la proyección de los puntos resultantes. GRS80 para coordenadas geográficas y MERC para coordenadas Spherical Mercator. Por default será GRS80. 
        -----------

        Para más información consultar: https://www.inegi.org.mx/servicios/Ruteo/Default.html#token

        '''
        return self.__obtener_ruta(linea_inicial, linea_final, destino_inicial, destino_final, tipo_vehiculo, ruta, ejes_excedentes, saltar_lineas, proyeccion, 'detalle_')


    def Combustibles(self): 
        '''
        Regresa un DataFrame con los 4 tipos de combustibles más comunes y su costo promedio que se consultan el primer día hábil de cada semana en la página web de la Comisión Reguladora de Energía del Gobierno Federal: https://www.gob.mx/cre

        Nota: el dato que provee esta API solo es una referencia en función del precio promedio nacional excluyendo las 7 regiones sobre la frontera. También lo que refiere al gas LP el precio es un promedio ponderado que publica la Comisión Reguladora de Energía.

        Para más información consultar: https://www.inegi.org.mx/servicios/Ruteo/Default.html#token

        '''
        params = {'key':self.__token, 
                    'type': 'json'}
        return self.__obtener_consulta('combustible', params, as_geodf = False)


import pandas as pd
import requests
import json
import warnings
from numpy import nan

class Indicadores:
    
    def __init__(self, token):
        self.__token = token # token proporcionado por el INEGI único para cada usuario
        self.__liga_base = 'https://www.inegi.org.mx/app/api/indicadores/desarrolladores/jsonxml/INDICATOR/'
        # vriables para la consulta
        self._indicadores = list() # lista con los indicadores a consultar. cada módulo la llena con sus especificaciones
        self._bancos = list() # lista con los bancos de los indicadores
        self._columnas = list() # nombres de las columnas. En los módulos de series se llena automáticamente.
        # diccionario con la clave de frecuencia del INEGI y el factor por el cual se debe multiplicar
        # el último valor para pasarlo a su mes correspondiente
        # Ejemplo: una serie semestral tiene como clave de frecuencia '4', esto indica que para cada año van
        # a haber dos periodos indicando los dos semestres: "2020/01, 2020/02"
        # El factor del diccionario indicaría que al último dígito lo multiplicamos por 6 para pasarlo a meses
        # las claves que no se encuentran en el diccionario son irregulares y no se van a operar
        self.__frecuancias_dict = {'BIE': {'1':(1,'Y'), '2':(1,'Y'), '3':(1,'Y'), '4':(6,'M'), '5':(4,'M'), '6':(3,'Q'), 
                                            '7':(2,'M'), '8':(1,'M'), '9':(14,'SM')},
                                    'BISE': {'1':(1,'Y'), '3':(1,'Y'), '4':(3,'Q'), '7':(14,'SM'), '8':(1,'M'), '9':(1,'Y'), '16':(1,'Y')}}
        self.__clave_entidad = None
        
############## Obtener Data Frame ########################

# Se definen  los métodos internos y públicos necesarios para obtener la serie y pasarla a un DataFrame
# Las dos variables esenciales para esto son self._indicadores y self._bancos. Cada módulo debe contar con métodos
# o atributos que permitan definir estas variables. También todas las variables de consulta deben ser accesibles tanto 
# dentro de las funciones obtener_df() y grafica() como parámetros así como atributos de la clase. 

    # aquí falta un control de errores cuando no se pudo obtener la info y advirtiendo que se cheque bien el token
    def __request(self, liga_api):
        req = requests.get(liga_api)
        assert req.status_code == 200, 'No se encontró información con las parámetros especificados.'
        data = json.loads(req.text)
        return data

    def __obtener_json(self, indicador, banco):
        """ 
    
        Construye la liga de consulta y obtiene la información resultante del API del
        INEGI. 
        
        Parámetros:
        ------------
        
        indicador: str.  Número del indicador a buscar.
        
        banco: str. Banco de información donde se encuentra el indocador. Puede ser 
                    'BIE' o 'BISE'.
        
       -------------
    
        Regresa un diccionario con la información del indicador proporcionada por el API
        del INEGI en formato JSON. 
    
        Para más información visitar https://www.inegi.org.mx/servicios/api_indicadores.html
    
        """
        if banco: 
            liga_api = '{}{}/es/{}/false/{}/2.0/{}?type=json'.format(self.__liga_base, indicador, self.__clave_entidad, banco, str(self.__token))
            data = self.__request(liga_api)
        else:
            try: 
                banco = 'BIE'
                liga_api = '{}{}/es/{}/false/{}/2.0/{}?type=json'.format(self.__liga_base, indicador, self.__clave_entidad, banco, str(self.__token))
                data = self.__request(liga_api)
            except: 
                banco = 'BISE'
                liga_api = '{}{}/es/{}/false/{}/2.0/{}?type=json'.format(self.__liga_base, indicador,  self.__clave_entidad, banco, str(self.__token))
                data = self.__request(liga_api)

        return data, banco
    
    def __json_a_df(self, data, banco):
        """ 
        
        Construye un DataFrame con la información resultante del API del INEGI de 
        un solo indicador a la vez.
    
        Parámetros:
        -----------
        data: Dict. JSON obtendio del API del INEGI.
        banco: str. Banco de información donde se encuentra el indocador. Puede ser 
                    'BIE' o 'BISE'.
        -----------
        
        Regresa un objeto tipo DataFrame con la información del indicador proporcionada por el API
        del INEGI en formato JSON. 
    
        Para más información visitar https://www.inegi.org.mx/servicios/api_indicadores.html
    
        """
        obs_totales = len(data['Series'][0]['OBSERVATIONS'])
        dic = {'fechas':[data['Series'][0]['OBSERVATIONS'][i]['TIME_PERIOD'] for i in range(obs_totales)],
                'valor':[float(data['Series'][0]['OBSERVATIONS'][i]['OBS_VALUE']) if data['Series'][0]['OBSERVATIONS'][i]['OBS_VALUE'] is not None else nan for i in range(obs_totales)]}
        df = pd.DataFrame.from_dict(dic)
        frecuencia = data['Series'][0]['FREQ']
        factor, period = self.__frecuancias_dict[banco].get(frecuencia) # factor que multiplica el periodo para pasar a fecha y periodo de pd
        if factor: 
            try: 
                cambio_fechas = lambda x: '/'.join(x.split('/')[:-1] + [str(int(x.split('/')[-1])*factor)])
                df.fechas = df.fechas.apply(cambio_fechas)
                df.set_index(pd.to_datetime(df.fechas),inplace=True, drop=True)
                df = df.drop(['fechas'],axis=1)
                if period == 'SM': df.index = df.index + pd.offsets.SemiMonthBegin()
            except: 
                df.set_index(df.fechas,inplace=True, drop=True)
                df = df.drop(['fechas'],axis=1)
        else:
            df.set_index(df.fechas,inplace=True, drop=True)
            df = df.drop(['fechas'],axis=1)
        return df

    def __definir_cve_ent(self, entidad):
        cve_base = '0700'
        if entidad == '00': 
            self.__clave_entidad = cve_base
        if len(entidad[2:5]) == 0: self.__clave_entidad = '{}00{}'.format(cve_base, entidad[:2])
        else: self.__clave_entidad = '{}00{}0{}'.format(cve_base, entidad[:2], entidad[2:5])

    # en este módulo solo se definen inicio y fin como variables ya que son las únicas que necesariamente están en
    # cada módulo. El resto de las variables se van definiendo en los módulos que hereden esta clase. 
    def _consulta(self, inicio, fin):
        
        if isinstance(self._indicadores, str): self._indicadores = [self._indicadores]
        if isinstance(self._bancos, str): self._bancos = [self._bancos]
        if isinstance(self._columnas, str): self._columnas = [self._columnas]
        
        lista_df = []
        for i in range(len(self._indicadores)):
            indicador = self._indicadores[i]
            try: banco = self._bancos[i]
            except: banco = None
            data, banco = self.__obtener_json(indicador, banco)
            df = self.__json_a_df(data, banco)
            if banco == 'BIE': df = df[::-1]
            lista_df.append(df)

        df = pd.concat(lista_df,axis=1)
        try: df.columns = self._columnas
        except: warnings.warn('Los nombres no coinciden con el número de indicadores')

        return df[inicio:fin] 
        
    # Las variables de la consulta pueden ser definidos o redefinidos con los parámteros de la función
    def obtener_df(self, 
                   indicadores: 'str|list', 
                   nombres: 'str|list' = None, 
                   clave_area: str = '00',
                   inicio: str = None, 
                   fin: str = None):
        """
        Regresa un DataFrame con la información de los indicadores proporcionada por el API del INEGI.
        Parametros
        -----------
        indicadores: str/list. Lista con los indicadores de las series a obtener.
        nombres: list/str, opcional. Lista con los nombres de las columas del DataFrame. De no proporcionarse, se usarán los indicadores.
        clave_area: str. Clave de dos a cinco caracteres que indentifica el área geográfica de acuerdo con el Marco Geoestadístico. Para definir el total nacional se especifica '00'. Este campo solo aplica para los indicadores del Bando de Indicadores (BISE), no aplica para los del Banco de Información Económica (BIE).
                                    Dos dígitos para incluir nivel estatal (ej.01 a 32).
                                    Cinco dígitos dígitos para incluir nivel municipal (ej. 01001).s
        inicio: str, opcional. Fecha donde iniciar la serie. De no proporcionarse será desde el primer valor disponible. 
        fin: str, opcional. Fecha donde terminar la serie. De no proporcionarse será hasta el último valor disponible. 
        ----------
        
        El DataFrame resultante tiene una columna por cada indicador y un DateTimeIndex con la fecha de los valores. 
        
        Para más información visitar https://www.inegi.org.mx/servicios/api_indicadores.html
        """     
        self._indicadores = indicadores
        self._columnas = indicadores
        if nombres: 
            self._columnas = nombres
        self.__definir_cve_ent(clave_area)
        return self._consulta(inicio, fin)




















##################### Graficar #######################

# Esta función queda fuera porque simplemente no le terminpe por ver sentido. dejo el código por si acaso.

# define los métodos públicos e internos para elaborar una gráfica base con la serie consultada
# por default utiliza la última consulta generada o definida aunque se puede redefinir variables de la consulta 
# dentro de la función.

#    def __cambiar_lineas(self, ax, estilo):
 #       if estilo == 'color':
  #          self.__cambiar_colores(ax)
   #     if estilo == 'ByN':
    #        self.__cambiar_estilos(ax)
     #   else: raise Exception('El estilo no coincide. Solo existe "color" y "ByN".')
      #  ax.legend(self._columnas)

 #   def __cambiar_estilos(self, ax):
  #      # checar cuál es la mejor manera de elejir los estilos
   #     line_styles = ['-','-.','--',':']
    #    markers = ['o','v','s','x','D']
     #   for i, line in enumerate(ax.get_lines()):
      #      if i <= 3: line.set_linestyle(line_styles[i])
       #     else: line.set_marker(markers[i-4])

 #   def __cambiar_colores(self, ax):
  #      palette = sns.color_palette('colorblind',len(self._columnas))[::-1]
   #     for i, line in enumerate(ax.get_lines()):
    #        line.set_color(palette[i])

 #   def grafica(self, estilo: str = 'color', show: bool = True, filename: str = None, **kwargs):
        """
        Construye un gráfico con la consulta definida. En caso de querer cambiar la consulta se pueden indicar los parámetros deseados: inicio, fin, indicador, serie, o cualquiera de los parámetros particulares de la serie.

        NOTA:
        Esta función no pretende remplazar el uso de librerías especializadas como Matplotlib o Seaborn, sino automatizar estilos de gráficas que puedan ser de uso común. Por ello, la gráfica generada tiene solo ciertos estilos disponibles. 
 #       Para darle un estilo particular o agregar nuevos elementos es recomendado usar alguna de las librerías especializadas directamente con los datos o #para manipular los objetos fig y ax que regresa esta función. 
#
 #       Parámetros
#        -----------
    #    estilo -- str. Estilo de la gráfica. Actualmente solo existen dos estilos: "color" y "ByN".
   #     show -- bool. Define si mostrar o no la gráfica. Equivalente a plt.show()
  #      filename -- nombre y dirección donde guardar la gráfica. Equivalente a plt.savefig()
 #       -----------
#
 #       Regresa los objetos fig y ax de matplotlib con la gráfica generada. 
#
#        """
#        df = self.obtener_df(**kwargs)
#        color = 'blue' if estilo == 'color' else 'black'
#        fig, ax = plt.subplots()
#        df.plot(ax=ax,color=color)
#        ax.legend().remove()
#        if len(self._df.columns)>1: self.__cambiar_lineas(ax, estilo)
#        sns.despine()
#        ax.set_xlabel('')
#        ax.ticklabel_format(style='plain',axis='y')
#        plt.tight_layout()
        
  #      if show: plt.show()
 #       if filename: plt.savefig(filename)
 #       return fig, ax



# -*- coding: utf-8 -*-
from setuptools import setup
import pathlib

HERE = pathlib.Path(__file__).parent
README = (HERE / "long_desc.md").read_text(encoding='utf-8')


NAME = 'INEGIpy'
DESCRIPTION = 'Librería en Python para facilitar el uso de las APIs del INEGI'
URL = 'https://github.com/andreslomeliv/DatosMex/tree/master/INEGIpy'
EMAIL = 'andres.lomeli.v@gmail.com'
AUTHOR = 'Andres Lomelí Viramontes'
REQUIRES_PYTHON = '>=3.6.0'
VERSION = '1.0.4'

# librerías requeridas
REQUIRED = [
		'requests','pandas','numpy','geopandas','shapely'
		]

setup(
	name=NAME,
	version=VERSION,
	description=DESCRIPTION,
	long_description=README,
	author=AUTHOR,
	author_email=EMAIL,
	python_requires=REQUIRES_PYTHON,
	url=URL,
	packages=['INEGIpy'],
	install_requires=REQUIRED,
	license='MIT'
)


import pandas as pd
import requests
import json
import geopandas as gpd
from shapely.geometry import Point

class DENUE:
    def __init__(self, token):
        self.__token = token
        self.__liga_base = 'https://www.inegi.org.mx/app/api/denue/v1/consulta/'
    
    # realiza las transformaciones necesarias para hacer iun geodataframe
    def __obtener_geodataframe(self, df): 
        df[['Longitud','Latitud']] = df[['Longitud','Latitud']].astype(float)
        df['geometry'] = df[['Longitud','Latitud']].values.tolist()
        df.geometry = df.geometry.apply(Point)
        geo_df = gpd.GeoDataFrame(df)
        geo_df.crs = 'EPSG:4326'
        return geo_df
     
    #ahorita parece que no vale la pena esta funcion pero tal vez sirva más en el futuro
    #su punto es estandarizar los strings para de nombres o ramas economicas etc
    # al final parece que esta función sí es totalmente inútil
    def __estadarizar_string(self, s): 
        s = s.lower().strip()
        return s
    
    # metodo general para obtener el json y pasarlo a un df o geodf
    def __obtener_consulta(self, liga, as_geodf):
        req = requests.get(liga)
        assert req.status_code == 200, 'No se encontró información con las parámetros especificados. \n Liga: {}'.format(liga)
        data = json.loads(req.text)
        df = pd.DataFrame(data)
        if as_geodf: df = self.__obtener_geodataframe(df)
        if len(df.columns) == 3: df.loc[:,'Total'] = df.loc[:, "Total"].astype(int)
        return df
        
    def Buscar(self, 
               condiciones: 'str|lsit', 
               latitud: float, 
               longitud: float, 
               distancia: int, 
               as_geodf: bool = True):
        ''' 
        Realiza una consulta de todos los establecimientos que cumplan las condiciones definidas.
        Esta función permite obtener la información de los establecimientos registrados en el DENUE dentro de un área definida de acuerdo con una lista de condiciones.
        
        Parámetros:
        ------------
        condiciones: str/list. Palabra(s) a buscar en el nombre del establecimiento, razón social, calle, colonia, clase de la actividad económica, entidad federativa, municipio y localidad. Para buscar todos los establecimientos se deberá ingresar la palabra "todos".
        latitud: float. Latitud que define el punto en el mapa a partir del cual se hará la consulta alrededor.
        longitud: float. Longitud que define el punto en el mapa a partir del cual se hará la consulta alrededor.
        distancia: int. Distancia en metros a partir de las coordenadas que definen el radio de búsqueda. La distancia máxima es de 5 000 metros.
        as_geodf: bool. Si el valor es verdadero regresa un GeoDataFrame para facilitar el análisis espacial. 
        ------------
        
        Regresa un DataFrame o GeoDataFrame con la información de la consulta.
        
        Para más información ir a: https://www.inegi.org.mx/servicios/api_denue.html
        
        '''
        if isinstance(condiciones, str): condiciones = [condiciones]
        condiciones = [self.__estadarizar_string(s) for s in condiciones] # checar si es necesario estas dos lineas
        condiciones = ','.join(condiciones)
        coordenadas = '{},{}'.format(str(latitud),str(longitud))
        liga = self.__liga_base + 'Buscar/{}/{}/{}/{}'.format(condiciones, coordenadas, str(distancia), self.__token)
        return self.__obtener_consulta(liga, as_geodf)
        
    
    def Ficha(self, 
              clave: str, 
              as_geodf: bool = True):
        ''' 
        Obtiene la información de un establecimiento en específico.
        
        Parámetros:
        ------------
        clave: str. ID única del establecimiento.
        as_geodf: bool. Si el valor es verdadero regresa un GeoDataFrame para facilitar el análisis espacial.
        ------------
        
        Regresa un DataFrame o GeoDataFrame con la información de la consulta.

        Para más información ir a: https://www.inegi.org.mx/servicios/api_denue.html
        
        '''
        liga = self.__liga_base + 'Ficha/{}/{}'.format(clave, self.__token)
        return self.__obtener_consulta(liga, as_geodf)
    
    def Nombre(self, 
               nombre: str, 
               entidad: str = '00', 
               registro_inicial: int = 1, 
               registro_final: int = 10, 
               as_geodf: bool = True):
        ''' 
        Realiza una consulta de todos los establecimientos por nombre o razón social.
        
        Parámetros:
        ------------
        nombre: str. Nombre del establecimiento ó razón social.
        entidad: str. Clave de dos dígitos de la entidad federativa (01 a 32). Para incluir todas las entidades se especifica 00.
        registro_inicial: int. Número de registro a partir del cuál se mostrarán los resultados de la búsqueda.
        registro_final: int. Número de registro final que se mostrará en los resultados de la búsqueda.
        as_geodf: bool. Si el valor es verdadero regresa un GeoDataFrame para facilitar el análisis espacial.
        ------------
        
        Regresa un DataFrame o GeoDataFrame con la información de la consulta.

        Para más información ir a: https://www.inegi.org.mx/servicios/api_denue.html
        
        '''
        nombre = self.__estadarizar_string(nombre)
        liga = self.__liga_base + 'Nombre/{}/{}/{}/{}/{}'.format(nombre, entidad, str(registro_inicial), str(registro_final), self.__token)
        return self.__obtener_consulta(liga, as_geodf)
    
    def BuscarEntidad(self, 
                      condiciones: 'str|list', 
                      entidad: str = '00', 
                      registro_inicial: int = 1, 
                      registro_final: int = 10, 
                      as_geodf: bool = True):
        ''' 
        Realiza una consulta de todos los establecimientos que cumplan las condiciones definidas y puede ser acotada por entidad federativa.
        
        Parámetros:
        ------------
        condiciones: str/list. Palabra(s) a buscar en el nombre del establecimiento, razón social, calle, colonia, clase de la actividad económica, entidad federativa, municipio y localidad. Para buscar todos los establecimientos se deberá ingresar la palabra "todos".
        entidad: str. Clave de dos dígitos de la entidad federativa (01 a 32). Para incluir todas las entidades se especifica 00.
        registro_inicial: int. Número de registro a partir del cuál se mostrarán los resultados de la búsqueda.
        registro_final: int. Número de registro final que se mostrará en los resultados de la búsqueda.
        as_geodf: bool. Si el valor es verdadero regresa un GeoDataFrame para facilitar el análisis espacial.
        ------------
        
        Regresa un DataFrame o GeoDataFrame con la información de la consulta.

        Para más información ir a: https://www.inegi.org.mx/servicios/api_denue.html
        
        '''
        if isinstance(condiciones, str): condiciones = [condiciones]
        condiciones = [self.__estadarizar_string(s) for s in condiciones] # checar si es necesario estas dos lineas
        condiciones = ','.join(condiciones)
        liga = self.__liga_base + 'BuscarEntidad/{}/{}/{}/{}/{}'.format(condiciones, entidad, str(registro_inicial), str(registro_final), self.__token)
        return self.__obtener_consulta(liga, as_geodf)
    
    # sirve para pasar de claves concatenadas de area y actividad económica a su versión separada para la consulta
    # evita que se llenen un montón de parámteros y simplifica las siguientes funciones
    def __desconcatenador(self, clave_area, clave_actividad):
        # para clave de area
        entidad = clave_area[:2] if len(clave_area[:2]) != 0 else '00'
        municipio = clave_area[2:5] if len(clave_area[2:5]) != 0 else '0'
        localidad = clave_area[5:9] if len(clave_area[5:9]) != 0 else '0'
        ageb = clave_area[9:13] if len(clave_area[9:13]) != 0 else '0'
        manzana = clave_area[13:16] if len( clave_area[13:16]) != 0 else '0'
        
        # para sector
        sector = clave_actividad if len(clave_actividad) == 2 else '0'
        subsector = clave_actividad if len(clave_actividad) == 3 else '0'
        rama = clave_actividad if len(clave_actividad) == 4 else '0'
        clase = clave_actividad if len(clave_actividad) == 6 else '0'
        
        return (entidad, municipio, localidad, ageb, manzana), (sector, subsector, rama, clase)

    def BuscarAreaAct(self, 
                      nombre: str, 
                      clave_area: str = '', 
                      clave_actividad: str = '', 
                      registro_inicial: int = 1, 
                      registro_final: int = 10,  
                      clave_establecimiento: str = '0', 
                      estrato: str = '0', 
                      as_geodf: bool = True):
        ''' 
        Realiza una consulta de todos los establecimientos con la opción de acotar la búsqueda por área geográfica, actividad económica, nombre, clave del establecimiento y estrato económico.
        
        Parámetros:
        ------------
        nombre: str. Nombre del establecimiento a buscar.
        clave_area: str. Clave de dos a dieciseis caracteres que indentifica el área geográfica de acuerdo con el Marco Geoestadístico. En caso de no definir una clave se regresan todos los establecimientos del país.
                                    Dos caracteres para incluir nivel estatal (ej. 01 a 32).
                                    Cinco caracteres dígitos para incluir nivel municipal (ej. 01001).
                                    Nueve caracteres para incluir nivel localidad (ej. 010010001).
                                    Trece caracteres para incluir nivel Área Geoestadística Básica (AGEB) (ej. 010010001216A).
                                    Dieciseis caracteres para incluir nivel manzana (ej. 010010001465A004).
        clave_actividad: str. Clave de dos a seis dígitos que identifica el área de actividad económica del establecimiento de acuerdo con el Sistema de Clasificación Industrial de América del Norte 2018. En caso de no definir una clave se regresan todas las áreas.
                                    Dos dígitos para definir un sector (ej.46).
                                    Tres dígitos para definir un subsector (ej. 464).
                                    Cuatro dígitos para definir una rama (ej. 4641).
                                    Cinco dígitos para definir una subrama (ej. 46411).
                                    Seis dígitos para definir una clase (ej. 464111).
        registro_inicial: int. Número de registro a partir del cuál se mostrarán los resultados de la búsqueda.
        registro_final: int. Número de registro final que se mostrará en los resultados de la búsqueda.
        clave_establecimiento: str. Clave única del establecimiento. Para incluir todos los establecimientos se especifica 0.
        estrato: str. Clave de un dígito que identifica el estrato del establecimiento (cantidad de trabajadores). Para incluir todos los tamaños se especifica 0.
                                    1. Para incluir de 0 a 5 personas.
                                    2. Para incluir de 6 a 10 personas.
                                    3. Para incluir de 11 a 30 personas.
                                    4. Para incluir de 31 a 50 personas.
                                    5. Para incluir de 51 a 100 personas.
                                    6. Para incluir de 101 a 250 personas.
                                    7. Para incluir de 251 y más personas.
        as_geodf: bool. Si el valor es verdadero regresa un GeoDataFrame para facilitar el análisis espacial.
        ------------
        
        Regresa un DataFrame o GeoDataFrame con la información de la consulta.

        Para más información ir a: https://www.inegi.org.mx/servicios/api_denue.html
        
        Para consultar las claves de actividad económica ir a: https://www.inegi.org.mx/app/scian/
        
        '''
        cves_area, cves_act = self.__desconcatenador(clave_area, clave_actividad)
            
        nombre = self.__estadarizar_string(nombre)
            
        liga = self.__liga_base + 'BuscarAreaActEstr/{}/{}/{}/{}/{}/{}/{}/{}/{}/{}/{}/{}/{}/{}/{}'.format(cves_area[0],cves_area[1], cves_area[2], cves_area[3], cves_area[4],
                                                                                                          cves_act[0], cves_act[1], cves_act[2], cves_act[3],
                                                                                                          nombre, str(registro_inicial), str(registro_final),
                                                                                                          clave_establecimiento, estrato, self.__token)
        return self.__obtener_consulta(liga, as_geodf)

    def Cuantificar(self, 
                    clave_area: 'str|list' = '0',
                    clave_actividad: 'str|list' = '0', 
                    estrato: str = '0'):
        ''' 
        Realiza un conteo de todos los establecimientos con la opción de acotar la búsqueda por área geográfica, actividad económica y estrato.
        
        A diferencia de las otras funciones permite múltiples claves en cada campo.
        
        Otra diferencia importante es que el resultado de esta consulta no regresa valores de las coordenadas para obtener un GeoDataFrame.
        
        Parámetros:
        ------------
        clave_area: str/list. Clave(s) de dos a nueve caracteres que indentifica el área geográfica de acuerdo con el Marco Geoestadístico. En caso de no definir una clave se regresan todos los establecimientos del país.
                                    Dos dígitos para incluir nivel estatal (ej.01 a 32).
                                    Cinco dígitos dígitos para incluir nivel municipal (ej. 01001).
                                    Nueve dígitos para incluir nivel localidad (ej. 010010001).
        clave_actividad: str/list. Clave(s) de dos a seis dígitos que identifican el área de actividad económica de los establecimientos de acuerdo con el Sistema de Clasificación Industrial de América del Norte 2018. Para incluir todas las actividades se especifica 0.
                                    Dos dígitos para definir un sector (ej.46).
                                    Tres dígitos para definir un subsector (ej. 464).
                                    Cuatro dígitos para definir una rama (ej. 4641).
                                    Cinco dígitos para definir una subrama (ej. 46411).
                                    Seis dígitos para definir una clase (ej. 464111).
        estrato: str. Clave de un dígito que identifica el estrato del establecimiento (cantidad de trabajadores). Para incluir todos los tamaños se especifica 0.
                                    1. Para incluir de 0 a 5 personas.
                                    2. Para incluir de 6 a 10 personas.
                                    3. Para incluir de 11 a 30 personas.
                                    4. Para incluir de 31 a 50 personas.
                                    5. Para incluir de 51 a 100 personas.
                                    6. Para incluir de 101 a 250 personas.
                                    7. Para incluir de 251 y más personas.
        ------------
        
        Regresa un DataFrame con la información de la consulta.

        Para más información ir a: https://www.inegi.org.mx/servicios/api_denue.html
        
        '''
        if isinstance(clave_actividad, list): clave_actividad = ','.join(clave_actividad)
        if isinstance(clave_area, list): clave_area = ','.join(clave_area)
        liga = self.__liga_base + 'Cuantificar/{}/{}/{}/{}'.format(clave_actividad, clave_area, estrato, self.__token)
        return self.__obtener_consulta(liga, False)
        